from mitama.app import Controller
from mitama.app.http import Response
from mitama.models import User, Group, InnerRole, Node, is_admin
from .model import Repo, Merge, InnerPermission
from .forms import MergeCreateForm, SettingsForm, HookCreateForm
from . import gitHttpBackend

import git
import os
import glob
import shutil
import yaml
import traceback
from io import StringIO
from unidiff import PatchSet


class RepoController(Controller):
    def handle(self, request):
        template = self.view.get_template("repo/list.html")
        repos = Repo.list()
        return Response.render(template, {
            'repos': repos
        })
    def create(self, request):
        template = self.view.get_template("repo/create.html")
        nodes = [
            *Group.list(),
            *User.list()
        ]
        try:
            if request.method == 'POST':
                body = request.post()
                repo = Repo()
                repo.name = body['name']
                repo.owner = Node.retrieve(body['owner'])
                repo.create()
                if not (self.app.project_dir / 'git_template').is_dir:
                    git.Repo.init(
                        self.app.project_dir / 'git_template',
                        bare = True
                    )
                git.Repo.init(
                    self.app.project_dir / ('repos/' + repo.name + '.git'),
                    bare = True,
                    template = self.app.project_dir / 'git_template'
                )
                return Response.redirect(self.app.convert_url('/'+repo.name))
        except Exception as err:
            error = str(err)
            traceback.print_exc(err)
            return Response.render(template, {
                'post': body,
                'error': error,
                'nodes': nodes
            })
        return Response.render(template, {
            'post': dict(),
            'nodes': nodes
        })

    def update(self, request):
        template = self.view.get_template("repo/update.html")
        repo = Repo.retrieve(name = request.params['repo'])
        if isinstance(repo.owner.object, Group) and not InnerPermission.is_accepted('update_repository', repo.owner.object, request.user):
            return Response.redirect(self.app.convert_url('/'))
        try:
            if request.method == 'POST':
                body = request.post()
                name = repo.name
                repo.name = body['name']
                repo.owner = Node(body['owner'])
                repo.update()
                os.rename(
                    self.app.project_dir / ('repos/' + name + '.git'),
                    self.app.project_dir / ('repos/' + repo.name + '.git')
                )
        except Exception as err:
            error = str(err)
            return Response.render(template, {
                'repo': repo,
                'error': error
            })
        return Response.render(template, {
            'repo': repo
        })

    def delete(self, request):
        template = self.view.get_template("repo/delete.html")
        repo = Repo.retrieve(name = request.params['repo'])
        if isinstance(repo.owner.object, Group) and not InnerPermission.is_accepted('delete_repository', repo.owner.object, request.user):
            return Response.redirect(self.app.convert_url('/'))
        try:
            if request.method == 'POST':
                if not request.user.password_check(request.post()['password']):
                    raise AuthorizationError('wrong password')
                shutil.rmtree(self.app.project_dir / ('repos/' + repo.name + '.git'))
                repo.delete()
                return Response.redirect(self.app.convert_url('/'))
        except Exception as err:
            error = str(err)
            return Response.render(template, {
                'repo': repo,
                'error': error
            })
        return Response.render(template, {
            'repo': repo
        })

    def retrieve(self, request):
        template = self.view.get_template("repo/retrieve.html")
        repo = Repo.retrieve(name = request.params['repo'])
        query = request.query
        current_head = request.params.get('head', 'master')
        entity = git.Repo(
            self.app.project_dir / 'repos/{}.git'.format(repo.name),
        )
        head = getattr(entity.heads, current_head) if hasattr(entity.heads, current_head) else None
        commit = None
        tree = None
        readme = None
        if head:
            commit = head.commit
            if commit:
                tree = commit.tree
                if tree:
                    for blob in tree:
                        if blob.name.startswith('README'):
                            readme = blob.data_stream.read().decode('utf-8')
        return Response.render(template, {
            'repo': repo,
            'current_head': current_head,
            'head': head,
            'tree': tree,
            'entity': entity,
            'readme': readme
        })

    def blob(self, request):
        template = self.view.get_template("repo/blob.html")
        repo = Repo.retrieve(name = request.params['repo'])
        query = request.query
        branch = query.get('branch', 'master')
        entity = git.Repo(
            self.app.project_dir / 'repos/{}.git'.format(repo.name),
        )
        head = getattr(entity.heads, branch) if hasattr(entity.heads, branch) else None
        tree = head.commit.tree or None
        content = None
        for obj in tree:
            if obj.name == request.params['object']:
                content = obj.data_stream.read().decode("utf-8")
        return Response.render(template, {
            'repo': repo,
            'branch': branch,
            'head': head,
            'tree': tree,
            'entity': entity,
            'name': request.params['object'],
            'content': content
        })

    def commit(self, request):
        template = self.view.get_template("repo/commit.html")
        repo = Repo.retrieve(name = request.params['repo'])
        query = request.query
        entity = repo.entity
        commit = entity.commit(request.params['commit'])
        diff_str = entity.git.diff(str(commit) + '~1', commit, ignore_blank_lines=True, ignore_space_at_eol=True) if len(commit.parents) > 0 else None
        diff = None
        if diff_str:
            diff = PatchSet(diff_str)
            for patch in diff:
                for hunk in patch:
                    print(dir(hunk))
        return Response.render(template, {
            'repo': repo,
            'entity': entity,
            'commit': commit,
            'diff': diff
        })

    def log(self, request):
        template = self.view.get_template("repo/log.html")
        repo = Repo.retrieve(name = request.params['repo'])
        current_head = request.params.get('head', 'master')
        query = request.query
        commits = repo.entity.iter_commits(current_head) if hasattr(repo.entity.heads, current_head) else None
        return Response.render(template, {
            'repo': repo,
            'entity': repo.entity,
            'current_head': current_head,
            'commits': commits
        })

class HookController(Controller):
    def handle(self, request):
        template = self.view.get_template('hook/list.html')
        repo = Repo.retrieve(name = request.params['repo'])
        hooks = list()
        for hook in glob.glob(str(self.app.project_dir / 'repos/{}.git/hooks'.format(repo.name)) + '/*'):
            hooks.append(os.path.basename(hook))
        return Response.render(template, {
            'repo': repo,
            'hooks': hooks
        })
    def retrieve(self, request):
        template = self.view.get_template('hook/retrieve.html')
        repo = Repo.retrieve(name = request.params['repo'])
        with open(self.app.project_dir / 'repos/{}.git/hooks/{}'.format(repo.name, request.params['hook'])) as f:
            code = f.read()
        return Response.render(template, {
            'repo': repo,
            'name': request.params['hook'],
            'code': code
        })
    def create(self, request):
        template = self.view.get_template('hook/create.html')
        repo = Repo.retrieve(name = request.params['repo'])
        if isinstance(repo.owner.object, Group) and not InnerPermission.is_accepted('update_repository', repo.owner.object, request.user):
            return Response.redirect(self.app.convert_url('/' + request.params['repo'] + '/hook'))
        error = ''
        if request.method == 'POST':
            form = HookCreateForm(request.post())
            with open(self.app.project_dir / 'repos/{}.git/hooks/{}'.format(repo.name, form['name']), 'w') as f:
                f.write(form['code'])
            return Response.redirect(self.app.convert_url('/{}/hook/{}'.format(repo.name, form['name'])))
        return Response.render(template, {
            'repo': repo,
            'error': error
        })
    def update(self, request):
        template = self.view.get_template('hook/update.html')
        repo = Repo.retrieve(name = request.params['repo'])
        if isinstance(repo.owner.object, Group) and not InnerPermission.is_accepted('update_repository', repo.owner.object, request.user):
            return Response.redirect(self.app.convert_url('/' + request.params['repo'] + '/hook'))
        error = ''
        if request.method == 'POST':
            form = HookCreateForm(request.post())
            shutil.move(
                self.app.project_dir / 'repos/{}.git/hooks/{}'.format(repo.name, request.params['hook']),
                self.app.project_dir / 'repos/{}.git/hooks/{}'.format(repo.name, form['name'])
            )
            with open(self.app.project_dir / 'repos/{}.git/hooks/{}'.format(repo.name, form['name']), 'w') as f:
                f.write(form['code'])
            return Response.redirect(self.app.convert_url('/{}/hook/{}'.format(repo.name, form['name'])))
        with open(self.app.project_dir / 'repos/{}.git/hooks/{}'.format(repo.name, request.params['hook'])) as f:
            code = f.read()
        return Response.render(template, {
            'repo': repo,
            'name': request.params['hook'],
            'code': code,
            'error': error
        })
    def delete(self, request):
        template = self.view.get_template('hook/delete.html')
        repo = Repo.retrieve(name = request.params['repo'])
        if isinstance(repo.owner.object, Group) and not InnerPermission.is_accepted('update_repository', repo.owner.object, request.user):
            return Response.redirect(self.app.convert_url('/' + request.params['repo'] + '/hook'))
        error = ''
        if request.method == 'POST':
            os.remove(self.app.project_dir / 'repos/{}.git/hooks/{}'.format(repo.name, request.params['hook']))
            return Response.redirect(self.app.convert_url('/{}/hook'.format(repo.name)))
        return Response.render(template, {
            'repo': repo,
            'name': request.params['hook'],
            'error': error
        })

class MergeController(Controller):
    def handle(self, request):
        template = self.view.get_template('merge/list.html')
        repo = Repo.retrieve(name = request.params['repo'])
        merges = Merge.query.filter(Merge.repo == repo).all()
        return Response.render(template, {
            "repo": repo,
            "merges": merges
        })

    def create(self, request):
        template = self.view.get_template('merge/create.html')
        repo = Repo.retrieve(name = request.params['repo'])
        error = ""
        if request.method == "POST":
            try:
                form = MergeCreateForm(request.post())
                merge = Merge()
                merge.base = form['base']
                merge.compare = form['compare']
                merge.title = form['title']
                merge.body = form['body']
                merge.repo = repo
                merge.user = request.user
                merge.create()
                return Response.redirect(self.app.convert_url('/' + repo.name + '/merge/' + merge._id))
            except Exception as err:
                error = str(err)
        return Response.render(template, {
            "repo": repo,
            "entity": repo.entity,
            "error": error
        })

    def retrieve(self, request):
        template = self.view.get_template('merge/retrieve.html')
        repo = Repo.retrieve(name = request.params['repo'])
        merge = Merge.retrieve(request.params['merge'])
        if request.method == 'POST':
            post = request.post()
            if post['action'] == 'merge':
                merge.merge()
                merge.delete()
            else:
                merge.delete()
            return Response.redirect(self.app.convert_url('/' + repo.name))
        return Response.render(template, {
            "repo": repo,
            "entity": repo.entity,
            "merge": merge
        })

class ProxyController(Controller):
    def handle(self, request):
        repo = Repo.retrieve(name = request.params['repo'][:-4])
        if repo.owner.object._id != request.user._id and (isinstance(repo.owner, Group) and not repo.owner.object.is_in(request.user)):
            return Response(status=401, reason='Unauthorized', text='You are not the owner of the repository.')
        environ = dict(request.environ)
        environ['REQUEST_METHOD'] = request.method
        environ['PATH_INFO'] = self.app.revert_url(environ['PATH_INFO'])
        (
            status,
            reason,
            headers,
            body
        ) = gitHttpBackend.wsgi_to_git_http_backend(environ, self.app.project_dir / 'repos')
        content_type = headers['Content-Type']
        return Response(
            body = body,
            status = status,
            reason = reason,
            headers = headers,
            content_type = content_type
        )

class SettingController(Controller):
    def handle(self, request):
        if not is_admin(request.user):
            return Response.redirect(self.app.convert_url('/'))
        error = ""
        if request.method == "POST":
            try:
                form = SettingsForm(request.post())
                for permission_screen_name, permission_roles in form['permission'].items():
                    permission = InnerPermission.retrieve(screen_name = permission_screen_name)
                    permission.roles = [InnerRole.retrieve(screen_name = role) for role in permission_roles]
                    permission.update()
                error = "変更を保存しました"
            except Exception as err:
                error = str(err)
        template = self.view.get_template('settings.html')
        return Response.render(template, {
            "roles": InnerRole.list(),
            "permissions": InnerPermission.list(),
            "error": error
        })





from mitama.db import BaseDatabase, relationship
from mitama.db.types import *
from mitama.models import inner_permission, Role, User, Node
import git
import hashlib
import shutil
import asyncio


class Database(BaseDatabase):
    pass


db = Database(prefix="izanami")

class Repo(db.Model):
    name = Column(String(64), primary_key=True, unique=True)
    owner = relationship(Node)
    owner_id = Column(String(64), ForeignKey("mitama_node._id"))
    @property
    def entity(self):
        entity = git.Repo(
            self.project_dir / 'repos/{}.git'.format(self.name),
        )
        return entity
    def merge(self, source, target):
        dirname = hashlib.sha256()
        dirname.update(self.name.encode())
        dirname.update(source.encode())
        dirname.update(target.encode())
        dirname = dirname.hexdigest()
        repo = git.Repo.clone_from(
            self.project_dir / 'repos/{}.git'.format(self.name),
            self.project_dir / 'tmp/{}'.format(dirname),
            branch=source
        )
        repo.index.merge_tree('origin/' + target).commit("Merged into {}".format(source))
        repo.remotes.origin.push()
        shutil.rmtree(self.project_dir / 'tmp/{}'.format(dirname))

class Merge(db.Model):
    repo_id = Column(String(64), ForeignKey("izanami_repo._id"))
    repo = relationship(Repo)
    base = Column(String(255), nullable=False)
    compare = Column(String(255), nullable=False)
    body = Column(String(1000))
    title = Column(String(255))
    user_id = Column(String(64), ForeignKey("mitama_user._id"))
    user = relationship(User)
    @property
    def meta(self):
        md = markdown.Metadata(extensions=["meta"])
        data = md.convert(self.body)
        return data.Meta

    def merge(self):
        self.repo.merge(self.base, self.compare)
        self.on("merge")()

Merge.listen("merge")

InnerPermission = inner_permission(db, [
    {
        "name": "ブランチのマージ",
        "screen_name": "merge"
    },
    {
        "name": "masterへのpush",
        "screen_name": "push_master",
    },
    {
        "name": "developへのpush",
        "screen_name": "push_develop",
    },
    {
        "name": "その他ブランチへのpush",
        "screen_name": "push_other",
    },
    {
        "name": "リポジトリの作成",
        "screen_name": "create_repository"
    },
    {
        "name": "リポジトリの削除",
        "screen_name": "delete_repository"
    },
    {
        "name": "リポジトリの設定",
        "screen_name": "update_repository"
    }
])

db.create_all()



from mitama.app import App, Router
from mitama.utils.controllers import static_files
from mitama.utils.middlewares import BasicMiddleware, SessionMiddleware, CsrfMiddleware
from mitama.app.method import view

#from .controller import RepoController, ProxyController, HookController
from .controller import RepoController, ProxyController, HookController, MergeController, SettingController
from .model import Repo, Merge, InnerPermission


class App(App):
    name = 'Izanami'
    description = 'Git server for Mitama.'
    merge = [Repo, Merge]
    router = Router(
        [
            view("/static/<path:path>", static_files()),
            Router([
                view("/<repo:re:(.*)\.git><path:path>", ProxyController),
            ], middlewares = [BasicMiddleware]),
            Router([
                view("/", RepoController),
                view("/create", RepoController, 'create'),
                view("/settings", SettingController),
                view("/<repo>", RepoController, 'retrieve'),
                view("/<repo>/update", RepoController, 'update'),
                view("/<repo>/delete", RepoController, 'delete'),
                view("/<repo>/tree/<head>", RepoController, 'retrieve'),
                view("/<repo>/blob/<head>/<object>", RepoController, 'blob'),
                view("/<repo>/commit/<commit>", RepoController, 'commit'),
                view("/<repo>/log", RepoController, 'log'),
                view("/<repo>/log/<head>", RepoController, 'log'),
                view("/<repo>/merge", MergeController),
                view("/<repo>/merge/create", MergeController, 'create'),
                view("/<repo>/merge/<merge>", MergeController, 'retrieve'),
                view("/<repo>/hook", HookController),
                view("/<repo>/hook/create", HookController, 'create'),
                view("/<repo>/hook/<hook>", HookController, 'retrieve'),
                view("/<repo>/hook/<hook>/edit", HookController, 'update'),
                view("/<repo>/hook/<hook>/delete", HookController, 'delete'),
            ], middlewares = [SessionMiddleware, CsrfMiddleware])
        ]
    )

    def init_app(self):
        Repo.project_dir = self.project_dir

    @property
    def view(self):
        view = super().view
        view.globals.update(
            permission=InnerPermission.is_accepted,
        )
        return view



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['izanami']

package_data = \
{'': ['*'],
 'izanami': ['static/*',
             'static/highlight/*',
             'templates/*',
             'templates/action/*',
             'templates/hook/*',
             'templates/merge/*',
             'templates/repo/*']}

install_requires = \
['GitPython>=3.1.13,<4.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'docker>=4.4.4,<5.0.0',
 'fabric>=2.6.0,<3.0.0',
 'invoke>=1.5.0,<2.0.0',
 'mitama>=4.3.3,<5.0.0',
 'unidiff>=0.6.0,<0.7.0']

setup_kwargs = {
    'name': 'izanami',
    'version': '1.1.0',
    'description': '',
    'long_description': None,
    'author': 'boke0',
    'author_email': 'speken00.tt@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from mitama.app.forms import Form, Field, DictField

class MergeCreateForm(Form):
    title = Field(label='要約', required=True)
    body = Field(label='メッセージ', required=True)
    base = Field(label='ベース', required=True)
    compare = Field(label='マージ対象', required=True)

class SettingsForm(Form):
    permission = DictField(listed=True)

class HookCreateForm(Form):
    name = Field(label="フック名", required=True)
    code = Field(label="コード")



"""Utility functions to invoke git-http-backend
"""

import subprocess
import threading


DEFAULT_CHUNK_SIZE = 0x8000
DEFAULT_MAX_HEADER_SIZE = 0X20000  # No header should ever be this large.
# TODO: expose these sizes to API
CRLF = b'\r\n'
HEADER_END = CRLF * 2


def wsgi_to_git_http_backend(wsgi_environ,
                             git_project_root,
                             user=None):
    """Convenience wrapper for how a WSGI application can use this
    module to handle a request.

    See build_cgi_environ regarding git_project_root and user.

    See run_git_http_backend for requirements for wsgi.input
    and wsgi.errors."""
    cgi_environ = build_cgi_environ(wsgi_environ, git_project_root, user)
    input_stream = wsgi_environ['wsgi.input']
    error_stream = wsgi_environ['wsgi.errors']
    cgi_header, response_body_generator = run_git_http_backend(
        cgi_environ, input_stream, error_stream
    )
    status_line, list_of_headers = parse_cgi_header(cgi_header)
    status, reason = status_line.split(' ')
    return status, reason, list_of_headers, response_body_generator


def run_git_http_backend(cgi_environ, input_stream, error_stream):
    """Execute "git http-backend" as a CGI script, using the supplied
    environment and the file-like object input_stream.

    See build_cgi_environ() and git documentation for the requirements
    for cgi_environ .

    input_stream can be any object implementing the file protocol. Note
    that input_stream will be closed here.

    Any stderr generated by the git process will be piped to error_stream,
    which must have a file descriptor.

    Return (cgi_header, response_body_generator). The cgi_header is the
    string of raw headers returned by git ending with just one CRLF. The
    response sent back to the client will need an additional blank line
    separating this from the response body.

    Raise EnvironmentError (errno 1) if a CGI/HTTP header is not returned
    from git http-backend."""
    input_length = int(cgi_environ.get('CONTENT_LENGTH', '') or 0)
    proc = subprocess.Popen(
        ['git', 'http-backend'],
        bufsize=DEFAULT_CHUNK_SIZE,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=error_stream,
        env=cgi_environ
    )
    cgi_header, response_body_generator = _communicate_with_git(
        proc, input_stream, input_length
    )
    return cgi_header, response_body_generator


def build_cgi_environ(wsgi_environ, git_project_root, user=None):
    """Build a CGI environ from a WSGI environment:

    CONTENT_TYPE
    GIT_PROJECT_ROOT = directory containing bare repos
    PATH_INFO (if GIT_PROJECT_ROOT is set, otherwise PATH_TRANSLATED)
    QUERY_STRING
    REMOTE_USER
    REMOTE_ADDR
    REQUEST_METHOD

    The git_project_root parameter must point to a directory that contains
    the git bare repo designated by PATH_INFO. See the git documentation.

    The git repo (my-repo.git) is located at GIT_PROJECT_ROOT + PATH_INFO
    (if GIT_PROJECT_ROOT is defined) or at PATH_TRANSLATED.

    If REMOTE_USER is set in wsgi_environ, you should normally leave user
    alone.
    """
    cgi_environ_ = dict(wsgi_environ)
    cgi_environ = dict()
    for key, value in cgi_environ_.items():  # NOT iteritems, due to "del"
        if isinstance(value, str):
            cgi_environ[key] = value
    cgi_environ['GIT_HTTP_EXPORT_ALL'] = '1'
    cgi_environ['GIT_PROJECT_ROOT'] = git_project_root
    if user:
        cgi_environ['REMOTE_USER'] = user
    cgi_environ.setdefault('REMOTE_USER', 'unknown')
    return cgi_environ


def parse_cgi_header(cgi_header):
    """Given the raw header returned by the CGI, return
    (status_line, list_of_headers). This adapts the CGI header
    to WSGI conventions."""
    header_dict = {}
    raw_lines = cgi_header.split(CRLF)
    assert raw_lines[-1] == b''
    for raw_line in raw_lines[:-1]:
        name, padded_value = raw_line.strip().split(b':', 1)
        value = padded_value.strip()
        header_dict[name.decode()] = value.decode()
    status_line = header_dict.pop('Status', None) or '200 OK'
    #list_of_headers = [(name, header_dict[name]) for name in names]
    #return status_line, list_of_headers
    return status_line, header_dict


def _communicate_with_git(proc, input_stream, input_length):
    # Given a subprocess.Popen object:
    # * Start writing request data
    # * Start reading stdout and possibly stderr
    # * Extract the cgi_header
    # * Construct a generator for everything that comes after the header
    # * Return (cgi_header, response_body_generator)
    # (The generator is responsible for extracting all data and cleaning up.)
    # Raise EnvironmentError (errno 1) if header is not returned from proc.
    threading.Thread(target=_input_data_pump,
                     args=(proc, input_stream, input_length)).start()
    chunks = [b'']  # Dummy str at start helps here.
    header_end = None
    while not header_end:
        total_bytes_read = sum(map(len, chunks))
        if total_bytes_read > DEFAULT_MAX_HEADER_SIZE:
            raise EnvironmentError(
                1,
                'Read %d bytes from "git http-backend" without '
                'finding header boundary.' % total_bytes_read,
            )  # TODO: Test this.
        chuck_data = proc.stdout.read(DEFAULT_CHUNK_SIZE)
        if not chuck_data:
            raise EnvironmentError(
                1,
                'Did not find header boundary in response '
                'from "git http-backend".',
            )  # TODO: Test this.
        chunks.append(chuck_data)
        # Search the two most recent chunks for the end of the header.
        # header_end -> (header_end_on_boundary, index_within_chunk) or None
        header_end = _find_header_end_in_2_chunks(*chunks[-2:])
    header_end_on_boundary, index_within_chunk = header_end
    cgi_header, remainder = _separate_header(
        chunks, header_end_on_boundary, index_within_chunk
    )
    response_body_generator = _response_body_generator(remainder, proc)
    return cgi_header, response_body_generator


def _input_data_pump(proc, input_stream, input_length):
    # Thread for feeding input to git
    # TODO: Currently using threads due to lack of universal standard for
    # async event loops in web applications.
    bytes_read = 0
    while bytes_read < input_length:
        bytes_to_read = min(DEFAULT_CHUNK_SIZE, input_length - bytes_read)
        current_data = input_stream.read(bytes_to_read)
        bytes_read += len(current_data)
        proc.stdin.write(current_data)
    proc.stdin.close()


def _find_header_end_in_2_chunks(chunk0, chunk1):
    # Search for the 4-byte HEADER_END in either the end of the first chunk
    # (with the 4-byte boundary stretching into the second chunk) or within
    # the second chunk starting at 0.
    # Return as (header_end_on_boundary, index_within_chunk).
    # Return None if header end not found.
    boundary_string = chunk0[-3:] + chunk1[:3]
    header_end = _search_str_for_header_end(boundary_string)
    if header_end != -1:
        return True, len(chunk0) - 3 + header_end
    header_end = _search_str_for_header_end(chunk1)
    if header_end != -1:
        return False, header_end
    return None


def _search_str_for_header_end(data_str):
    """Return index of header end or -1."""
    return data_str.find(HEADER_END)


def _separate_header(chunks, header_end_on_boundary, index_within_chunk):
    # Return header, remainder
    if header_end_on_boundary:
        # Header ends within chunks[-2]
        header_chunks = chunks[:-2]
        last_header_chunk = chunks[-2]
        body_start_index = (4 - (len(last_header_chunk) - index_within_chunk))
    else:
        # Header ends within chunks[-1]
        header_chunks = chunks[:-1]
        last_header_chunk = chunks[-1]
        body_start_index = index_within_chunk + 4
    header_chunks.append(last_header_chunk[:index_within_chunk])
    header_chunks.append(CRLF)  # Line end might have been split.
    header = b''.join(header_chunks)
    remainder = chunks[-1][body_start_index:]
    return header, remainder


def _response_body_generator(remainder, proc):
    # The generator returned up the stack to the WSGI application.
    # Yields chunks of data from the subprocess output.
    yield remainder
    current_data = proc.stdout.read(DEFAULT_CHUNK_SIZE)
    while current_data:
        yield current_data
        current_data = proc.stdout.read(DEFAULT_CHUNK_SIZE)
    # TODO: Do we need this?
    while proc.poll() is None:
        yield b''
